import multiprocessing as mp
import typing as t
from threading import Thread
from typeguard import typechecked

class MultiWorker:

    _sentinel = None
    _finished_adding = False

    @typechecked
    def __init__(self, job: callable, init: callable = None, init_args: t.Optional[tuple] = None, worker_count: int = mp.cpu_count()) -> None:
        self._tasks = mp.Queue()
        self._results = mp.Queue()
        self._workers = []
        for _ in range(worker_count):
            self._workers.append(mp.Process(target = MultiWorker._worker, args = (job, init, init_args, self._tasks, self._results)))
        self._overlord = Thread(target = MultiWorker._overlord, args = (self._workers, self._tasks, self._results))

    @staticmethod
    @typechecked
    def _worker(job: callable, init: callable, init_args: t.Optional[tuple], tasks: mp.Queue, results: mp.Queue) -> None:
        state = None
        if init is not None:
            state = init(init_args)
        while True:
            item = tasks.get()
            if item == MultiWorker._sentinel:
                tasks.put(MultiWorker._sentinel)
                break            
            else:
                if state is None:
                    result = job(item)
                else:
                    result = job(state, item)
                results.put(result)

    @staticmethod
    @typechecked
    def _overlord(workers: t.List[mp.Process], tasks: mp.Queue, results: mp.Queue) -> None:
        for worker in workers:
            worker.join()
        results.put(MultiWorker._sentinel)
        tasks.close()

    @typechecked
    def start(self) -> None:
        for worker in self._workers:
            worker.start()
        self._overlord.start()
    
    @typechecked
    def add_task(self, item) -> None:
        self._tasks.put(item)

    @typechecked
    def finished_adding_tasks(self) -> None:
        if not self._finished_adding:
            self._finished_adding = True
            self._tasks.put(MultiWorker._sentinel)

    @typechecked
    def get_results(self) -> t.Iterator:
        while True:
            item = self._results.get()
            if item == MultiWorker._sentinel:
                break
            else:
                yield item
        self._results.close()




import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name = "mp_boilerplate",
    version = "0.0.1",
    author = '@markanewman',
    author_email = 'mp_boilerplate@trinetteandmark.com',
    description = 'A collection of patterns to use over top of the built in multiprocessing package',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = "https://github.com/markanewman/mp_boilerplate",
    project_urls = {
        'Bug Reports': 'https://github.com/markanewman/mp_boilerplate/issues',
        'Source': 'https://github.com/markanewman/mp_boilerplate',
    },
    packages = setuptools.find_packages(),
    classifiers = [        
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ],
    python_requires = '>=3.6, <4',
    install_requires = [
        "typeguard==2.7.1"]
)